import base58
import hashlib
import mimetypes
import os
import urllib.parse
from typing import Any

from django.db import models
from django.db.models.fields.files import FieldFile, FileField
from django.utils.functional import cached_property

from anchor.tasks import blob_make_variant

DEFAULT_CONTENT_TYPE = "application/octet-stream"

# Add extra known types
mimetypes.add_type("image/jpeg", ".jfif")
mimetypes.add_type("image/webp", ".webp")


class VariantFieldFile(FieldFile):
    """
    A replacement for Django's FieldFile that works with BlobField to support
    variant generation.
    """

    def get_variant(self, blocking=False, **format_params):
        """
        Returns a variant of this file with the transformations specified by the
        keyword arguments.

        If the transformation was previously calculated, it is retrieved from
        cache.
        """

        # calculate unique name for the given parameters
        variant_name = self.get_variant_filename(**format_params)

        if self.storage.exists(variant_name):
            return self.storage.open(variant_name)

        if blocking:
            return self.make_variant(variant_name, format_params)

        blob_make_variant(self, variant_name, format_params)

    def get_variant_url(self, fallback=False, blocking=False, **format_params):
        """
        Returns the url of a variant of this file according to the params
        specified in format_params. The file is not opened if it already exists
        for efficiency. If the variant does not exist, it is created per
        `self.get_variant(**format_params)`.

        If fallback=True and the variant does not exists, the original file URL
        (which is guaranteed to be available) is returned so that the image can
        be rendered immediately.
        """
        variant_name = self.get_variant_filename(**format_params)
        try:
            if not self.storage.exists(variant_name):
                # ensure variant starts processing
                self.get_variant(blocking=blocking, **format_params)

                if blocking:
                    # if blocking the URL will be available
                    return self.storage.url(variant_name)
                elif fallback:
                    #  return a URL that is valid now if fallback is enabled
                    return self.url
                else:
                    return

        except NotImplementedError:  # pragma: no cover
            try:
                return self.url
            except ValueError:
                return None

        return self.storage.url(variant_name)

    def make_variant(self, variant_name, format_params):
        """Makes a variant synchronously."""

        # not cached, open file for writing
        variant_file = self.storage.open(variant_name, mode="wb")

        # process according to type
        if self.is_image:
            from anchor.transformers import PillowImageTransformer

            transformer = PillowImageTransformer(**format_params)
            transformer.transform(self.file, variant_file)
            return self.storage.open(variant_name)
        else:
            raise NotImplementedError(
                "Could not find transformer for mime type %s" % self.mime_type
            )

    def get_variant_filename(self, **format_params):
        """
        Names a variant for a file by appending a hash of the parameters that
        produced it.
        """

        # calculate parameter dict hash
        h1 = urllib.parse.urlencode(sorted(format_params.items()))
        h2 = hashlib.sha1(h1.encode("utf8"))

        # encode to save characters in the filename
        suffix = base58.b58encode(h2.digest()).decode()[:-1]

        # append extension to help proxy servers identify the mime type
        _, ext = os.path.splitext(self.file.name)
        format = format_params.pop("format", ext[1:])
        return f"{self.name}_{suffix}.{format}"

    def get_fingerprint(self):
        return type(self).calculate_fingerprint(self)

    def get_dimensions(self):
        """
        Returns a tuple of integers (width, height) if this file is a raster
        image. Otherwise it returns (None, None).
        """
        if not self.is_image or self.is_vector:
            return None, None

        from PIL import Image

        image = Image.open(self.open())
        return image.size

    @property
    def is_image(self):
        return self.mime_type.startswith("image")

    @property
    def is_video(self):
        return self.mime_type.startswith("video")

    @property
    def is_vector(self):
        return "svg" in self.mime_type or "pdf" in self.mime_type

    @cached_property
    def mime_type(self):
        return type(self).guess_mime_type(self.name)

    @classmethod
    def guess_mime_type(cls, filename):
        """Attempts to guess the MIME type from the extension of the filename.

        Falls back to `DEFAULT_CONTENT_TYPE` (application/octet-stream) if the
        extension is not known."""
        if filename is None:
            return DEFAULT_CONTENT_TYPE

        mime, _ = mimetypes.guess_type(filename)

        if mime is not None:
            return mime

        return DEFAULT_CONTENT_TYPE

    @classmethod
    def calculate_fingerprint(cls, file):
        m = hashlib.sha1()
        file.seek(0)
        for c in file.chunks(1024):
            m.update(c)

        return m.hexdigest()

    @property
    def filename(self):
        return os.path.basename(self.name)

    @property
    def extension(self):
        return os.path.splitext(self.name)[1]


class VariantFileField(FileField):
    """
    A replacement for Django's FileField that works with VariantFieldFiles to
    support variant generation.
    """

    attr_class = VariantFieldFile


class BlobField(models.ForeignKey):
    """
    Attach a Blob to a model.

    This field is intended to replace the default Django FileField or ImageField
    in models.
    """

    def __init__(
        self,
        on_delete=models.SET_NULL,
        null=True,
        blank=False,
        related_name="+",
        **kwargs,
    ):
        to = "anchor.Blob"

        return super().__init__(
            to=to,
            on_delete=on_delete,
            related_name=related_name,
            null=null,
            blank=blank,
            **kwargs,
        )

    def formfield(self, **kwargs: Any) -> Any:
        from anchor.forms.fields import BlobField as BlobFormField

        return super().formfield(
            **{
                "form_class": BlobFormField,
                "required": not self.blank,
                **kwargs,
            }
        )

    def deconstruct(self) -> Any:
        name, path, args, kwargs = super().deconstruct()

        # remove "to" argument from kwargs
        kwargs.pop("to", None)

        return name, path, args, kwargs



from PIL import Image

from anchor.models.fields import VariantFieldFile


class BaseTransformer:
    def __init__(self, **format_params):
        self.format_params = format_params

        self._set_default_params()
        self._validate_format_params()

    def transform(self, src: VariantFieldFile, dst: VariantFieldFile):
        """
        Saves the transformed image to `variant_file`.
        """
        raise NotImplementedError

    def _set_default_params(self):
        """
        Assigns default parameters after initialization.
        """
        pass

    def _validate_format_params(self):
        """
        Called after self._set_default_format_params to validate their are
        consistent.

        Subclasses should override this method and raise ValueError with invalid
        parameters.
        """
        pass


class PillowImageTransformer(BaseTransformer):
    """
    Transformer for images using the Pillow package.

    Format parameters
    -----------------
    format : str or Tuple[str, Dict[str, Any]]
        Format in which to save the image. When this parameter is a tuple, the
        first argument is the format identifier, such as ``webp`` and the second
        is a dictionary of options. They are both used to call
        ``PIL.Image.save(format[0], **format[1])``. If the parameter is just a
        string, the second element of the tuple defaults to an empty dictionary
        so the Pillow defaults are used.

        See the Pillow image format documentation for more details at
        https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html

    thumbnail : Tuple[int, int]
        Resizes the image to fit inside a box of the provided dimensions using
        ``PIL.Image.thumbnail``. The resized version is never larger than the
        original image.

    """

    def _set_default_params(self):
        if "format" in self.format_params:
            # if the provided format is just a string, create an empty options
            # dictionary
            if isinstance(self.format_params["format"], str):
                self.format_params["format"] = (self.format_params["format"], {})

    def transform(self, src: VariantFieldFile, dst: VariantFieldFile):
        img = Image.open(src.open())

        # create thumbnail
        thumb_size = self.format_params.get("thumbnail")
        if thumb_size is not None:
            img.thumbnail(size=thumb_size)

        # convert to another format
        if "format" in self.format_params:
            format, options = self.format_params["format"]
        else:
            format = None
            options = {}

        img.save(dst.name, format=format, **options)

        return dst



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from anchor.models.base import BaseModel


class AttachmentQuerySet(models.QuerySet):
    def filter_by_object(self, object, name="attachments"):
        """
        Filter attachments by object and name.
        """
        content_type = ContentType.objects.get_for_model(object)
        return self.filter(content_type=content_type, object_id=object.pk, name=name)


class Attachment(BaseModel):
    class Meta:
        constraints = (
            models.constraints.UniqueConstraint(
                fields=("content_type", "object_id", "blob", "name"),
                name="unique_attachment_per_blob_and_object_and_name",
            ),
        )

    objects = AttachmentQuerySet.as_manager()

    blob = models.ForeignKey(
        "anchor.Blob",
        on_delete=models.PROTECT,
        related_name="attachments",
        null=True,
        blank=True,
        verbose_name="blob",
    )

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, verbose_name="content type"
    )
    object_id = models.CharField(db_index=True, max_length=22, verbose_name="object id")
    content_object = GenericForeignKey("content_type", "object_id")

    order = models.IntegerField(default=0, verbose_name="order")
    name = models.CharField(max_length=256, default="attachments", verbose_name="name")

    def __str__(self) -> str:
        return f"{self.name} {self.order}"



from django.contrib import admin
from django.contrib.contenttypes.admin import GenericTabularInline
from django.utils.html import format_html
from anchor.models.fields import BlobField
from anchor.forms.widgets import AdminBlobInput
from anchor.models.attachment import Attachment
from anchor.models.blob import Blob
from django.template.defaultfilters import filesizeformat


class BlobFieldMixin:
    """
    Render a preview of the blob in the admin form.

    Inherit from this in your ModelAdmin class to render a preview of the blob
    in the admin form.
    """

    formfield_overrides = {
        BlobField: {"widget": AdminBlobInput},
    }


@admin.register(Attachment)
class AttachmentAdmin(admin.ModelAdmin):
    list_display = ("blob", "name", "content_type", "content_object")
    raw_id_fields = ("blob",)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("blob", "content_type")
            .prefetch_related("content_object")
        )


@admin.register(Blob)
class BlobAdmin(BlobFieldMixin, admin.ModelAdmin):
    ordering = ("id",)
    date_hierarchy = "created_at"
    search_fields = ("filename", "id", "fingerprint", "uploaded_by__email")
    list_display = ("filename", "human_size", "uploaded_by", "created_at")
    list_filter = ("mime_type",)
    readonly_fields = ("filename", "byte_size", "fingerprint", "thumbnail")
    raw_id_fields = ("uploaded_by",)

    def save_model(self, request, obj, form, change):  # pragma: no cover
        if not change:
            obj.uploaded_by = request.user

        super().save_model(request, obj, form, change)

    @admin.display(description="Size", ordering="byte_size")
    def human_size(self, instance: Blob):
        return filesizeformat(instance.byte_size)

    def thumbnail(self, instance: Blob):
        if instance.file and instance.file.is_image:
            return format_html(
                '<img src="{}" style="max-width: 100%">', instance.file.url
            )

        return "-"


class AttachmentInline(GenericTabularInline):
    """
    Inline for Attachment model.

    Add this to the admin.ModelAdmin.inlines attribute of the model you want to attach files to.
    """

    model = Attachment
    extra = 0
    fields = ("blob", "name", "order", "thumbnail")
    readonly_fields = ("thumbnail",)
    ordering = ("name", "order")
    autocomplete_fields = ("blob",)

    def thumbnail(self, instance):
        if instance.blob:
            return format_html(
                '<img src="{}" style="max-width: 100%">', instance.blob.file.url
            )

        return "-"

    thumbnail.short_description = "Thumbnail"



# Generated by Django 5.0.7 on 2024-07-18 08:27

import anchor.models.base
import anchor.models.blob
import anchor.models.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Blob",
            fields=[
                (
                    "id",
                    models.CharField(
                        default=anchor.models.base._gen_short_uuid,
                        editable=False,
                        max_length=22,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created at"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated at"),
                ),
                (
                    "file",
                    anchor.models.fields.VariantFileField(
                        upload_to=anchor.models.blob._attachment_upload_to,
                        verbose_name="file",
                    ),
                ),
                (
                    "alt_text",
                    models.TextField(
                        blank=True,
                        help_text="Provide a description of the file for convenience and accessibility.",
                        null=True,
                        verbose_name="alternative text",
                    ),
                ),
                (
                    "filename",
                    models.CharField(
                        default=None,
                        max_length=256,
                        null=True,
                        verbose_name="original filename",
                    ),
                ),
                (
                    "byte_size",
                    models.PositiveBigIntegerField(
                        blank=True,
                        default=None,
                        help_text="size in bytes",
                        null=True,
                        verbose_name="size",
                    ),
                ),
                (
                    "fingerprint",
                    models.CharField(
                        db_index=True,
                        editable=False,
                        max_length=256,
                        null=True,
                        verbose_name="fingerprint",
                    ),
                ),
                (
                    "mime_type",
                    models.CharField(
                        default="application/octet-stream",
                        editable=False,
                        max_length=32,
                        verbose_name="MIME type",
                    ),
                ),
                (
                    "width",
                    models.PositiveSmallIntegerField(
                        blank=True, null=True, verbose_name="original width"
                    ),
                ),
                (
                    "height",
                    models.PositiveSmallIntegerField(
                        blank=True, null=True, verbose_name="original height"
                    ),
                ),
                (
                    "uploaded_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="uploaded by",
                    ),
                ),
            ],
            options={
                "verbose_name": "blob",
                "verbose_name_plural": "blobs",
            },
        ),
        migrations.CreateModel(
            name="Attachment",
            fields=[
                (
                    "id",
                    models.CharField(
                        default=anchor.models.base._gen_short_uuid,
                        editable=False,
                        max_length=22,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created at"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated at"),
                ),
                (
                    "object_id",
                    models.CharField(
                        db_index=True, max_length=22, verbose_name="object id"
                    ),
                ),
                ("order", models.IntegerField(default=0, verbose_name="order")),
                (
                    "name",
                    models.CharField(
                        default="attachments", max_length=256, verbose_name="name"
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                        verbose_name="content type",
                    ),
                ),
                (
                    "blob",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="attachments",
                        to="anchor.blob",
                        verbose_name="blob",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="attachment",
            constraint=models.UniqueConstraint(
                fields=("content_type", "object_id", "blob", "name"),
                name="unique_attachment_per_blob_and_object_and_name",
            ),
        ),
    ]



import uuid

import base58
from django.db import models
from django.utils.translation import gettext_lazy as _


def _gen_short_uuid():
    return base58.b58encode(uuid.uuid4().bytes).decode("ascii")


class BaseModel(models.Model):
    id = models.CharField(
        primary_key=True,
        max_length=22,
        verbose_name="ID",
        editable=False,
        default=_gen_short_uuid,
    )
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_("created at"))
    updated_at = models.DateTimeField(auto_now=True, verbose_name=_("updated at"))

    class Meta:
        abstract = True



import datetime
import logging
import os
import urllib

from django.conf import settings
from django.core.files import File as DjangoFile
from django.core.files.base import ContentFile
from django.db import models

from anchor.models.attachment import Attachment
from anchor.models.fields import (
    DEFAULT_CONTENT_TYPE,
    VariantFieldFile,
    VariantFileField,
)
from anchor.models.base import BaseModel

logger = logging.getLogger("anchor")


class BlobQuerySet(models.QuerySet):
    def create_and_attach(self, object=None, name=None, order=0, **kwargs):
        blob = super().create(**kwargs)
        Attachment.objects.create(
            blob=blob, content_object=object, order=order, name=name
        )
        return blob

    def from_url(self, url: str, **kwargs):
        """
        Downloads the file at the given URL and returns a Blob wrapping it.

        The Blob is persisted to the database but if there exists a blob with
        the same fingerprint, the existing one is reused.
        """

        with urllib.request.urlopen(url) as fp:
            contents = fp.read()
            blob = self.model.from_file(
                ContentFile(contents), filename=os.path.basename(url), **kwargs
            )
            blob.save()
            return blob

    def from_path(self, path: str, **kwargs):
        """
        Opens the file at the given ``path`` and returns a Blob wrapping it.

        The Blob is persisted to the database but if there exists a blob with
        the same fingerprint, the existing one is reused.
        """
        with open(path, mode="rb") as fp:
            blob = self.model.from_file(fp, **kwargs)
            blob.save()
            return blob


def _attachment_upload_to(instance, filename: str):
    filename = os.path.basename(filename)
    return datetime.datetime.now().strftime(
        str(f"anchor/blobs/%Y/%m/%d/{instance.pk}_{filename}")
    )


class Blob(BaseModel):
    class Meta:
        verbose_name = "blob"
        verbose_name_plural = "blobs"

    objects = BlobQuerySet.as_manager()

    file: VariantFieldFile = VariantFileField(
        upload_to=_attachment_upload_to, verbose_name="file"
    )
    alt_text = models.TextField(
        verbose_name="alternative text",
        help_text="Provide a description of the file for convenience and accessibility.",
        null=True,
        blank=True,
    )

    # automatic fields
    filename = models.CharField(
        max_length=256, verbose_name="original filename", null=True, default=None
    )
    byte_size = models.PositiveBigIntegerField(
        verbose_name="size",
        help_text="size in bytes",
        null=True,
        blank=True,
        default=None,
    )
    fingerprint = models.CharField(
        max_length=256,
        null=True,
        db_index=True,
        verbose_name="fingerprint",
        editable=False,
    )
    mime_type = models.CharField(
        max_length=32,
        default=DEFAULT_CONTENT_TYPE,
        verbose_name="MIME type",
        editable=False,
    )
    width = models.PositiveSmallIntegerField(
        null=True, blank=True, verbose_name="original width"
    )
    height = models.PositiveSmallIntegerField(
        null=True, blank=True, verbose_name="original height"
    )
    uploaded_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name="uploaded by",
    )

    def save(self, *args, **kwargs):
        if self.filename is None:
            try:
                self.filename = os.path.basename(self.file.name)
            except TypeError:  # pragma: no cover
                pass

        if self.byte_size is None:
            self.byte_size = self.file.size

        if self.fingerprint is None:
            self.fingerprint = self.file.get_fingerprint()

        if self.mime_type is None:
            self.mime_type = self.file.mime_type

        if self.width is None and self.height is None:
            self.width, self.height = self.file.get_dimensions()

        return super().save(*args, **kwargs)

    @classmethod
    def from_file(
        cls,
        file: DjangoFile | str,
        avoid_duplicates=True,
        filename=None,
        byte_size=None,
        **kwargs,
    ):
        """
        Initializes a Blob with the given arguments, making guesses for
        arguments not provided.

        If ``avoid_duplicates`` is True and a blob exists with the same hash, it
        will be fetched, updated and returned instead of creating a new blob for
        the current file.

        By default, returned blobs are not persisted to the database. Even if
        they were already present, new attributes passed via keyword arguments
        are only updated locally. To persist changes to the database pass the
        ``save=True`` keyword argument.

        Args:
            file (DjangoFile or str): The file to wrap in a Blob.
            avoid_duplicates (bool): Whether to reuse existing blobs with the same hash.
            filename (str): The name of the file.
            byte_size (int): The size of the file in bytes.
            kwargs: Additional fields to set on the Blob.
        """
        if not isinstance(file, DjangoFile):
            file = DjangoFile(file)

        if filename is None:
            try:
                filename = os.path.basename(file.name)
            except TypeError:  # pragma: no cover
                pass

        if byte_size is None:
            byte_size = file.size

        fingerprint = VariantFieldFile.calculate_fingerprint(file)

        kwargs.update(
            {
                "fingerprint": fingerprint,
                "mime_type": VariantFieldFile.guess_mime_type(filename),
                "byte_size": byte_size,
                "filename": filename,
            }
        )
        if avoid_duplicates:
            if existing := cls.objects.filter(fingerprint=fingerprint).first():
                for k, v in kwargs.items():
                    setattr(existing, k, v)
                try:
                    existing.file.open()
                except FileNotFoundError:
                    logger.warning(
                        "Found existing blob with matching fingerprint but missing file. Replacing file with given version."
                    )
                    existing.file.save(filename, file, save=True)

                return existing

        blob = cls(**kwargs)
        blob.file.save(filename, file, save=False)
        return blob

    def __str__(self):  # pragma: no cover
        return self.file.name



from django.forms.widgets import ClearableFileInput


class ClearableBlobInput(ClearableFileInput):
    """
    Replacement for Django's ClearableFileInput widget that works with
    BlobFields.
    """

    # make this widget work with the Django admin
    choices = []


class AdminBlobInput(ClearableBlobInput):
    template_name = "anchor/widgets/admin_blob_input.html"



from django import forms

from anchor.models.blob import Blob

from .widgets import ClearableBlobInput


class BlobField(forms.FileField):
    """
    A form field for uploading a file and storing it as a Blob.

    This field is intended to replace the default Django FileField in forms.
    """

    widget = ClearableBlobInput

    def __init__(self, *args, **kwargs):
        # remove ModelChoiceField attributes from kwargs for compatibility
        kwargs.pop("limit_choices_to", None)
        kwargs.pop("queryset", None)
        kwargs.pop("to_field_name", None)
        kwargs.pop("blank", None)
        return super().__init__(*args, **kwargs)

    def prepare_value(self, value: str):
        if value:
            try:
                blob = Blob.objects.get(pk=value)
                return blob.file
            except Blob.DoesNotExist:
                return None
        return value

    def clean(self, data, initial=None):
        if isinstance(initial, Blob):
            initial_file = initial.file
        elif isinstance(initial, str):
            initial_blob = Blob.objects.filter(pk=initial).first()
            if initial_blob:
                initial_file = initial_blob.file
            else:
                initial_file = None
        else:
            initial_file = None

        file = super().clean(data, initial=initial_file)

        if file is False:
            return None
        elif file is None:
            if initial is not None:
                return Blob.objects.get(pk=initial)
            else:
                return None
        else:
            blob = Blob.from_file(file)
            blob.save()
            return blob



import logging

from django import template

from anchor.models import Blob
from anchor.models.fields import VariantFieldFile

register = template.Library()
logger = logging.getLogger("attachments")


@register.simple_tag()
def blob(_blob: Blob | VariantFieldFile, blocking=False, **format_params):
    """
    Returns the file URL of a Blob.

    The URL returned corresponds to the original file that was uploaded as the
    blob (not any variant).

    You can also pass a VariantFieldFile instance to this tag, in which case it
    will be used instead of looking for the ``file`` attribute of a Blob. If the
    given object is neither a Blob or a VariantFieldFile, an empty string is
    returned.
    """
    if _blob is None:
        return ""

    if isinstance(_blob, Blob):
        blob_file = _blob.file
    elif isinstance(_blob, VariantFieldFile):
        blob_file = _blob
    else:
        return ""

    try:
        if format_params == {}:
            return blob_file.url
        return blob_file.get_variant_url(
            fallback=True, blocking=blocking, **format_params
        )
    except AttributeError:
        logger.exception("Error rendering blob variant.")
        return ""
    except FileNotFoundError:
        logger.error("Error opening blob variant file.")
        return ""


@register.simple_tag()
def blob_thumbnail(
    _blob: Blob | VariantFieldFile,
    max_width=1024,
    max_height=1024,
    format="webp",
    ignore_vectors=True,
    blocking=False,
    **format_params,
):
    """
    Returns the URL of a variant of the given Blob according to the specified
    format parameters.

    This tag is useful for rendering thumbnails or resized and optimized
    versions of image Blobs.
    """
    if isinstance(_blob, Blob):
        blob_file = _blob.file
    elif isinstance(_blob, VariantFieldFile):
        blob_file = _blob
    else:
        return ""

    if blob_file is None:
        return ""

    if ignore_vectors and blob_file.is_vector:
        return blob_file.url

    if blob_file.is_image:
        return blob(
            _blob,
            blocking=blocking,
            format=format,
            thumbnail=(max_width, max_height),
            **format_params,
        )

    return ""
